// dither.c (C) 2015 Claude Heiland-Allen <claude@mathr.co.uk>
// http://mathr.co.uk/blog/2015-02-24_dithering.html
// gcc -std=c99 -Wall -Wextra -pedantic -O3 -fopenmp -o dither dither.c -lfftw3 -lm
// ./dither
// generates 1.3GB of uncompressed image data

#include <complex.h>
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <fftw3.h>

const double pi = 3.141592653589793;

void hsv2rgb(double *hsv, double *rgb) {
  double h = hsv[0];
  double s = hsv[1];
  double v = hsv[2];
  double i, f, p, q, t, r, g, b;
  int ii;
  if (s == 0.0) { r = g = b = v; } else {
    h = 6 * (h - floor(h));
    ii = i = floor(h);
    f = h - i;
    p = v * (1 - s);
    q = v * (1 - (s * f));
    t = v * (1 - (s * (1 - f)));
    switch(ii) {
      case 0: r = v; g = t; b = p; break;
      case 1: r = q; g = v; b = p; break;
      case 2: r = p; g = v; b = t; break;
      case 3: r = p; g = q; b = v; break;
      case 4: r = t; g = p; b = v; break;
      default:r = v; g = p; b = q; break;
    }
  }
  rgb[0] = r;
  rgb[1] = g;
  rgb[2] = b;
}

// http://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.601-7-201103-I!!PDF-E.pdf
void rgb2yuv(double *rgb, double *yuv) {
  double r = rgb[0];
  double g = rgb[1];
  double b = rgb[2];
  double ey = 0.299 * r + 0.587 * g + 0.114 * b;
  double eu = (r - ey) / 1.402;
  double ev = (b - ey) / 1.772;
  yuv[0] = (219 * ey +  16) / 256;
  yuv[1] = (224 * eu + 128) / 256;
  yuv[2] = (224 * ev + 128) / 256;
}

void yuv2rgb(double *yuv, double *rgb) {
  double y = yuv[0];
  double u = yuv[1];
  double v = yuv[2];
  double ey = (256 * y - 16 ) / 219;
  double eu = (256 * u - 128) / 224;
  double ev = (256 * v - 128) / 224;
  rgb[0] = 1.402 * eu + ey;
  rgb[2] = 1.772 * ev + ey;
  rgb[1] = (ey - 0.299 * rgb[0] - 0.114 * rgb[2]) / 0.587;
}

double uniform() {
  return rand() / (double) RAND_MAX;
}

double normal() {
  double u1 = uniform();
  double u2 = uniform();
  double g1 = sqrt(-2 * log(u1)) * cos(2 * pi * u2);
//double g2 = sqrt(-2 * log(u1)) * sin(2 * pi * u2);
  if (isnan(g1) || isinf(g1)) {
    return normal();
  }
  return g1;
}

double dither(double x, int bits) {
  return x + normal() / (1 << bits);
}

int quantize(double x, int bits) {
  return fmin(fmax(floor(((1 << bits) - 1) * x + 0.5), 0), (1 << bits) - 1);
}

double unquantize(int x, int bits) {
  return x / (double) ((1 << bits) - 1);
}

int shift(int x, int b) {
  return b >= 0 ? x << b : x >> -b;
}

/*

example:

210
abc

->

76543210
abcabcab

*/
int extend(int x, int bits0, int bits1) {
  int y = 0;
  for (int b = bits1 - bits0; b > -bits0; b -= bits0) {
    y |= shift(x, b);
  }
  return y;
}

void save_ppm(char *filename, int width, int height, int bytes, void *pixels) {
  FILE *f = fopen(filename, "wb");
  fprintf(f, "P6\n%d %d\n%d\n", width, height, (1 << (8 * bytes)) - 1);
  fwrite(pixels, 3 * width * height * bytes, 1, f);
  fclose(f);
}

// for little-endian architecture
static inline uint16_t PPM16(uint16_t x) {
  return (x >> 8) | (x << 8);
}

/*
// for big-endian architecture
static inline uint16_t PPM16(uint16_t x) {
  return x;
}
*/

int main() {
  printf("BEGIN\n\n");
  FILE *csv = fopen("dither.csv", "wb");
  fprintf(csv, "#bits,mode,smi,sma,savg,sdev,ami,ama,aavg,adev\n");

  int width = 1024;
  int height = 1024;

  printf("planning FFT\n");
  fftw_complex *fft_in = fftw_malloc(sizeof(*fft_in) * width * height * 3);
  fftw_complex *fft_out = fftw_malloc(sizeof(*fft_out) * width * height * 3);
  fftw_plan plan[3];
  for (int c = 0; c < 3; ++c) {
    plan[c] = fftw_plan_dft_2d(height, width, fft_in + c * width * height, fft_out + c * width * height, FFTW_FORWARD, FFTW_MEASURE);
  }

  uint8_t  *ppm8  = malloc(3 * width * height * sizeof(*ppm8));
  uint16_t *ppm16 = malloc(3 * width * height * sizeof(*ppm16));
  double *image     = malloc(3 * width * height * sizeof(*image));
  int    *image_int = malloc(3 * width * height * sizeof(*image_int));
  double *image_err = malloc(3 * width * height * sizeof(*image_err));

#define LOOP(code) \
  for (int j = 0; j < height; ++j) { \
    for (int i = 0; i < width; ++i) { \
      for (int c = 0; c < 3; ++c) { \
        int k = 3 * (width * j + i) + c; \
        code \
      } \
    } \
  }

#define LOOP3(code) \
  for (int j = 0; j < height; ++j) { \
    for (int i = 0; i < width; ++i) { \
      int k = 3 * (width * j + i); \
      code \
    } \
  }

  printf("generating source RGB image\n");
  #pragma omp parallel for
  LOOP3(
    double x = (i + 0.5) / width - 0.5;
    double y = (j + 0.5) / height - 0.5;
    double r = sqrt(x * x + y * y) * 4;
    double hsv[3];
    hsv[0] = atan2(y, x) / (2 * pi);
    hsv[1] = tanh(r);
    hsv[2] = (1 + cos(pi * tanh(r * (5 + cos(pi * r + 5 * atan2(y, x)))/6)))/2;
    hsv2rgb(hsv, image + k);
  )

  char stem[100];
  char filename[100];

  for (int bits = 1; bits <= 16; ++bits) {

    printf("\nQUANTIZE %d bits\n\n", bits);

    for (int mode = 0; mode < 6; ++mode) {
      switch (mode) {
      case 0:
        snprintf(stem, 100, "B%02dDQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          // note: rand() might be non-deterministic
          #pragma omp parallel for
          LOOP(
            image_int[k] = quantize(dither(image[k], bits), bits);
            ppm8[k] = extend(image_int[k], bits, 8);
            image_err[k] = unquantize(image_int[k], bits) - image[k];
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          // note: rand() might be non-deterministic
          #pragma omp parallel for
          LOOP(
            image_int[k] = quantize(dither(image[k], bits), bits);
            ppm16[k] = PPM16(extend(image_int[k], bits, 16));
            image_err[k] = unquantize(image_int[k], bits) - image[k];
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      case 1:
        snprintf(stem, 100, "B%02dQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          #pragma omp parallel for
          LOOP(
            image_int[k] = quantize(image[k], bits);
            ppm8[k] = extend(image_int[k], bits, 8);
            image_err[k] = unquantize(image_int[k], bits) - image[k];
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          #pragma omp parallel for
          LOOP(
            image_int[k] = quantize(dither(image[k], bits), bits);
            ppm16[k] = PPM16(extend(image_int[k], bits, 16));
            image_err[k] = unquantize(image_int[k], bits) - image[k];
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      case 2:
        snprintf(stem, 100, "B%02dQYQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          #pragma omp parallel for
          LOOP3(
            double rgb[3];
            double yuv[3];
            for (int c = 0; c < 3; ++c) {
              rgb[c] = unquantize(image_int[k+c], bits);
            }
            rgb2yuv(rgb, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(yuv[c], bits);
              ppm8[k+c] = extend(image_int[k+c], bits, 8);
            }
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          #pragma omp parallel for
          LOOP3(
            double rgb[3];
            double yuv[3];
            for (int c = 0; c < 3; ++c) {
              rgb[c] = unquantize(image_int[k+c], bits);
            }
            rgb2yuv(rgb, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(yuv[c], bits);
              ppm16[k+c] = PPM16(extend(image_int[k+c], bits, 16));
            }
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      case 3:
        snprintf(stem, 100, "B%02dQYQRQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            double rgb[3];
            for (int c = 0; c < 3; ++c) {
              yuv[c] = unquantize(image_int[k+c], bits);
            }
            yuv2rgb(yuv, rgb);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(rgb[c], bits);
              ppm8[k+c] = extend(image_int[k+c], bits, 8);
              image_err[k+c] = unquantize(image_int[k+c], bits) - image[k+c];
            }
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            double rgb[3];
            for (int c = 0; c < 3; ++c) {
              yuv[c] = unquantize(image_int[k+c], bits);
            }
            yuv2rgb(yuv, rgb);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(rgb[c], bits);
              ppm16[k+c] = PPM16(extend(image_int[k+c], bits, 16));
              image_err[k+c] = unquantize(image_int[k+c], bits) - image[k+c];
            }
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      case 4:
        snprintf(stem, 100, "B%02dYQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(yuv[c], bits);
              ppm8[k+c] = extend(image_int[k+c], bits, 8);
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(yuv[c], bits);
              ppm16[k+c] = PPM16(extend(image_int[k+c], bits, 16));
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      case 5:
        snprintf(stem, 100, "B%02dYDQ", bits);
        snprintf(filename, 100, "%s.ppm", stem);
        printf("%s\n", filename);
        if (bits <= 8) {
          // note: rand() might be non-deterministic
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(dither(yuv[c], bits), bits);
              ppm8[k+c] = extend(image_int[k+c], bits, 8);
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 1, ppm8);
        } else if (bits <= 16) {
          // note: rand() might be non-deterministic
          #pragma omp parallel for
          LOOP3(
            double yuv[3];
            rgb2yuv(image + k, yuv);
            for (int c = 0; c < 3; ++c) {
              image_int[k+c] = quantize(dither(yuv[c], bits), bits);
              ppm16[k+c] = PPM16(extend(image_int[k+c], bits, 16));
              image_err[k+c] = unquantize(image_int[k+c], bits) - yuv[c];
            }
          )
          save_ppm(filename, width, height, 2, ppm16);
        }
        break;
      }

      snprintf(filename, 100, "%sA.ppm", stem);
      printf("%s\n", filename);
      int q = 8;
      #pragma omp parallel for schedule(static, 1)
      for (int y = 0; y < q; ++y) {
        for (int x = 0; x < q; ++x) {
          double rmi =  1.0/0.0;
          double rma = -1.0/0.0;
          for (int j = y * height/q; j < (y + 1) * height/q; ++j) {
            for (int i = x * width/q; i < (x + 1) * width/q; ++i) {
              for (int c = 0; c < 3; ++c) {
                int k = 3 * (j * width + i) + c;
                double uqu = unquantize(image_int[k], bits);
                rmi = fmin(rmi, uqu);
                rma = fmax(rma, uqu);
              }
            }
          }
          if (rma == rmi) {
            rma = rmi + 1;
          }
          for (int j = y * height/q; j < (y + 1) * height/q; ++j) {
            for (int i = x * width/q; i < (x + 1) * width/q; ++i) {
              for (int c = 0; c < 3; ++c) {
                int k = 3 * (j * width + i) + c;
                double uqu = unquantize(image_int[k], bits);
                ppm8[k] = quantize((uqu - rmi) / (rma - rmi), 8);
              }
            }
          }
        }
      }
      save_ppm(filename, width, height, 1, ppm8);

      snprintf(filename, 100, "%sE.ppm", stem);
      printf("%s\n", filename);
      double smi =  1.0/0.0;
      double sma = -1.0/0.0;
      double ss1 = 0.0;
      double ss2 = 0.0;
      double ami =  1.0/0.0;
      double ama = -1.0/0.0;
      double as1 = 0.0;
      double as2 = 0.0;
      // no parallel possible due to min/max reductions?
      // #pramga omp parallel for
      LOOP(
        double sx = image_err[k];
        double ax = fabs(sx);
        smi = fmin(smi, sx);
        sma = fmax(sma, sx);
        ss1 += sx;
        ss2 += sx * sx;
        ami = fmin(ami, ax);
        ama = fmax(ama, ax);
        as1 += ax;
        as2 += ax * ax;
      )
      double n = 3 * width * height;
      double savg = ss1 / n;
      double aavg = as1 / n;
      double sdev = sqrt(n * ss2 - ss1 * ss1) / n;
      double adev = sqrt(n * as2 - as1 * as1) / n;
      fprintf(csv, "%d,%d,%f,%f,%f,%f,%f,%f,%f,%f\n"
        , bits
        , mode
        , smi  * (1 << bits)
        , sma  * (1 << bits)
        , savg * (1 << bits)
        , sdev * (1 << bits)
        , ami  * (1 << bits)
        , ama  * (1 << bits)
        , aavg * (1 << bits)
        , adev * (1 << bits)
        );

      #pragma omp parallel for
      LOOP(
        image_err[k] /= ama;
        ppm8[k] = quantize(0.5 + 0.5 * image_err[k], 8);
        fft_in[c * width * height + j * width + i] = image_err[k];
      )
      save_ppm(filename, width, height, 1, ppm8);

      snprintf(filename, 100, "%sEF.ppm", stem);
      printf("%s\n", filename);
      #pragma omp parallel for schedule(static, 1)
      for (int c = 0; c < 3; ++c) {
        fftw_execute(plan[c]);
      }
      #pragma omp parallel for
      LOOP(
        int l = c * width * height + ((j + height/2)%height) * width + ((i + width/2)%width);
        image_err[k] = log(1 + cabs(fft_out[l]));
      )
      double fftma = 0;
      // no parallel possible due to max reduction?
      // #pragma omp parallel for
      LOOP(fftma = fmax(fftma, image_err[k]);)
      #pragma omp parallel for
      LOOP(ppm8[k] = quantize(image_err[k] / fftma, 8);)
      save_ppm(filename, width, height, 1, ppm8);

    } // mode
  } // bits

  fclose(csv);
  printf("\nEND\n");
  return 0;
}
